import click


@click.command(short_help='Show information about the current environment')
@click.pass_obj
def status(app):
    """Show information about the current environment."""

    def display_pair(key, value, display_func=None, link=None):
        app.display_info('[', end='')
        app.display_success(key, end='')
        app.display_info(']', end='')
        app.display_info(' - ', end='')
        (display_func or app.display_info)(value, link=link)

    if app.project.root is None:
        if app.project.chosen_name is None:
            display_pair('Project', '<no project detected>', app.display_warning)
        else:
            display_pair('Project', f'{app.project.chosen_name} (not a project)', app.display_warning)
    elif app.project.chosen_name is None:
        display_pair('Project', f'{app.project.root.name} (current directory)')
    else:
        display_pair('Project', app.project.chosen_name)

    display_pair('Location', str(app.project.location))
    display_pair('Config', str(app.config_file.path), link=f'file:///{app.config_file.path}')



from __future__ import annotations

import os

from hatch.config.constants import AppEnvVars


def get_env_var(*, plugin_name: str, option: str) -> str:
    return f'{AppEnvVars.ENV_OPTION_PREFIX}{plugin_name}_{option.replace("-", "_")}'.upper()


def get_env_var_option(*, plugin_name: str, option: str, default: str = '') -> str:
    return os.environ.get(get_env_var(plugin_name=plugin_name, option=option), default)


def ensure_valid_environment(env_config: dict):
    env_config.setdefault('type', 'virtual')


def get_verbosity_flag(verbosity: int, *, adjustment=0) -> str:
    verbosity += adjustment
    if not verbosity:
        return ''

    if verbosity > 0:
        return f'-{"v" * abs(min(verbosity, 3))}'

    return f'-{"q" * abs(max(verbosity, -3))}'


def add_verbosity_flag(command: list[str], verbosity: int, *, adjustment=0):
    flag = get_verbosity_flag(verbosity, adjustment=adjustment)
    if flag:
        command.append(flag)



import array
import errno
import logging
import socket
from contextlib import asynccontextmanager, contextmanager
from itertools import count
from typing import Optional

from outcome import Value, Error
import trio
from trio.abc import Channel

from jeepney.auth import Authenticator, BEGIN
from jeepney.bus import get_bus
from jeepney.fds import FileDescriptor, fds_buf_size
from jeepney.low_level import Parser, MessageType, Message
from jeepney.wrappers import ProxyBase, unwrap_msg
from jeepney.bus_messages import message_bus
from .common import (
    MessageFilters, FilterHandle, ReplyMatcher, RouterClosed, check_replyable,
)

log = logging.getLogger(__name__)

__all__ = [
    'open_dbus_connection',
    'open_dbus_router',
    'Proxy',
]


# The function below is copied from trio, which is under the MIT license:

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
@contextmanager
def _translate_socket_errors_to_stream_errors():
    try:
        yield
    except OSError as exc:
        if exc.errno in {errno.EBADF, errno.ENOTSOCK}:
            # EBADF on Unix, ENOTSOCK on Windows
            raise trio.ClosedResourceError("this socket was already closed") from None
        else:
            raise trio.BrokenResourceError(
                "socket connection broken: {}".format(exc)
            ) from exc



class DBusConnection(Channel):
    """A plain D-Bus connection with no matching of replies.

    This doesn't run any separate tasks: sending and receiving are done in
    the task that calls those methods. It's suitable for implementing servers:
    several worker tasks can receive requests and send replies.
    For a typical client pattern, see :class:`DBusRouter`.

    Implements trio's channel interface for Message objects.
    """
    def __init__(self, socket, enable_fds=False):
        self.socket = socket
        self.enable_fds = enable_fds
        self.parser = Parser()
        self.outgoing_serial = count(start=1)
        self.unique_name = None
        self.send_lock = trio.Lock()
        self.recv_lock = trio.Lock()
        self._leftover_to_send = None  # type: Optional[memoryview]

    async def send(self, message: Message, *, serial=None):
        """Serialise and send a :class:`~.Message` object"""
        async with self.send_lock:
            if serial is None:
                serial = next(self.outgoing_serial)
            fds = array.array('i') if self.enable_fds else None
            data = message.serialise(serial, fds=fds)
            await self._send_data(data, fds)

    # _send_data is copied & modified from trio's SocketStream.send_all() .
    # See above for the MIT license.
    async def _send_data(self, data: bytes, fds):
        if self.socket.did_shutdown_SHUT_WR:
            raise trio.ClosedResourceError("can't send data after sending EOF")

        with _translate_socket_errors_to_stream_errors():
            if self._leftover_to_send:
                # A previous message was partly sent - finish sending it now.
                await self._send_remainder(self._leftover_to_send)

            with memoryview(data) as data:
                if fds:
                    sent = await self.socket.sendmsg([data], [(
                        trio.socket.SOL_SOCKET, trio.socket.SCM_RIGHTS, fds
                    )])
                else:
                    sent = await self.socket.send(data)

                await self._send_remainder(data, sent)

    async def _send_remainder(self, data: memoryview, already_sent=0):
        try:
            while already_sent < len(data):
                with data[already_sent:] as remaining:
                    sent = await self.socket.send(remaining)
                already_sent += sent
            self._leftover_to_send = None
        except trio.Cancelled:
            # Sending cancelled mid-message. Keep track of the remaining data
            # so it can be sent before the next message, otherwise the next
            # message won't be recognised.
            self._leftover_to_send = data[already_sent:]
            raise

    async def receive(self) -> Message:
        """Return the next available message from the connection"""
        async with self.recv_lock:
            while True:
                msg = self.parser.get_next_message()
                if msg is not None:
                    return msg

                # Once data is read, it must be given to the parser with no
                # checkpoints (where the task could be cancelled).
                b, fds = await self._read_data()
                if not b:
                    raise trio.EndOfChannel("Socket closed at the other end")
                self.parser.add_data(b, fds)

    async def _read_data(self):
        if self.enable_fds:
            nbytes = self.parser.bytes_desired()
            with _translate_socket_errors_to_stream_errors():
                data, ancdata, flags, _ = await self.socket.recvmsg(
                    nbytes, fds_buf_size()
                )
            if flags & getattr(trio.socket, 'MSG_CTRUNC', 0):
                self._close()
                raise RuntimeError("Unable to receive all file descriptors")
            return data, FileDescriptor.from_ancdata(ancdata)

        else:  # not self.enable_fds
            with _translate_socket_errors_to_stream_errors():
                data = await self.socket.recv(4096)
            return data, []

    def _close(self):
        self.socket.close()
        self._leftover_to_send = None

    # Our closing is currently sync, but AsyncResource objects must have aclose
    async def aclose(self):
        """Close the D-Bus connection"""
        self._close()

    @asynccontextmanager
    async def router(self):
        """Temporarily wrap this connection as a :class:`DBusRouter`

        To be used like::

            async with conn.router() as req:
                reply = await req.send_and_get_reply(msg)

        While the router is running, you shouldn't use :meth:`receive`.
        Once the router is closed, you can use the plain connection again.
        """
        async with trio.open_nursery() as nursery:
            router = DBusRouter(self)
            await router.start(nursery)
            try:
                yield router
            finally:
                await router.aclose()


async def open_dbus_connection(bus='SESSION', *, enable_fds=False) -> DBusConnection:
    """Open a plain D-Bus connection

    :return: :class:`DBusConnection`
    """
    bus_addr = get_bus(bus)
    sock : trio.SocketStream = await trio.open_unix_socket(bus_addr)

    # Authentication
    authr = Authenticator(enable_fds=enable_fds, inc_null_byte=False)
    if hasattr(socket, 'SCM_CREDS'):
        # BSD: send credentials message to authenticate (kernel fills in data)
        await sock.socket.sendmsg(
            [b'\0'], [(socket.SOL_SOCKET, socket.SCM_CREDS, bytes(512))]
        )
    else:
        # Linux: no ancillary data needed, bus checks with SO_PEERCRED
        await sock.send_all(b'\0')
    for req_data in authr:
        await sock.send_all(req_data)
        authr.feed(await sock.receive_some())
    await sock.send_all(BEGIN)

    conn = DBusConnection(sock.socket, enable_fds=enable_fds)

    # Say *Hello* to the message bus - this must be the first message, and the
    # reply gives us our unique name.
    async with conn.router() as router:
        reply = await router.send_and_get_reply(message_bus.Hello())
        conn.unique_name = reply.body[0]

    return conn


class TrioFilterHandle(FilterHandle):
    def __init__(self, filters: MessageFilters, rule, send_chn, recv_chn):
        super().__init__(filters, rule, recv_chn)
        self.send_channel = send_chn

    @property
    def receive_channel(self):
        return self.queue

    async def aclose(self):
        self.close()
        await self.send_channel.aclose()

    async def __aenter__(self):
        return self.queue

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.aclose()


class Future:
    """A very simple Future for trio based on `trio.Event`."""
    def __init__(self):
        self._outcome = None
        self._event = trio.Event()

    def set_result(self, result):
        self._outcome = Value(result)
        self._event.set()

    def set_exception(self, exc):
        self._outcome = Error(exc)
        self._event.set()

    async def get(self):
        await self._event.wait()
        return self._outcome.unwrap()


class DBusRouter:
    """A client D-Bus connection which can wait for replies.

    This runs a separate receiver task and dispatches received messages.
    """
    _nursery_mgr = None
    _rcv_cancel_scope = None

    def __init__(self, conn: DBusConnection):
        self._conn = conn
        self._replies = ReplyMatcher()
        self._filters = MessageFilters()

    @property
    def unique_name(self):
        return self._conn.unique_name

    async def send(self, message, *, serial=None):
        """Send a message, don't wait for a reply
        """
        await self._conn.send(message, serial=serial)

    async def send_and_get_reply(self, message) -> Message:
        """Send a method call message and wait for the reply

        Returns the reply message (method return or error message type).
        """
        check_replyable(message)
        if self._rcv_cancel_scope is None:
            raise RouterClosed("This DBusRouter has stopped")

        serial = next(self._conn.outgoing_serial)

        with self._replies.catch(serial, Future()) as reply_fut:
            await self.send(message, serial=serial)
            return (await reply_fut.get())

    def filter(self, rule, *, channel: Optional[trio.MemorySendChannel]=None, bufsize=1):
        """Create a filter for incoming messages

        Usage::

            async with router.filter(rule) as receive_channel:
                matching_msg = await receive_channel.receive()

            # OR:
            send_chan, recv_chan = trio.open_memory_channel(1)
            async with router.filter(rule, channel=send_chan):
                matching_msg = await recv_chan.receive()

        If the channel fills up,
        The sending end of the channel is closed when leaving the ``async with``
        block, whether or not it was passed in.

        :param jeepney.MatchRule rule: Catch messages matching this rule
        :param trio.MemorySendChannel channel: Send matching messages here
        :param int bufsize: If no channel is passed in, create one with this size
        """
        if channel is None:
            channel, recv_channel = trio.open_memory_channel(bufsize)
        else:
            recv_channel = None
        return TrioFilterHandle(self._filters, rule, channel, recv_channel)

    # Task management -------------------------------------------

    async def start(self, nursery: trio.Nursery):
        if self._rcv_cancel_scope is not None:
            raise RuntimeError("DBusRouter receiver task is already running")
        self._rcv_cancel_scope = await nursery.start(self._receiver)

    async def aclose(self):
        """Stop the sender & receiver tasks"""
        # It doesn't matter if we receive a partial message - the connection
        # should ensure that whatever is received is fed to the parser.
        if self._rcv_cancel_scope is not None:
            self._rcv_cancel_scope.cancel()
            self._rcv_cancel_scope = None

        # Ensure trio checkpoint
        await trio.sleep(0)

    # Code to run in receiver task ------------------------------------

    def _dispatch(self, msg: Message):
        """Handle one received message"""
        if self._replies.dispatch(msg):
            return

        for filter in self._filters.matches(msg):
            try:
                filter.send_channel.send_nowait(msg)
            except trio.WouldBlock:
                pass

    async def _receiver(self, task_status=trio.TASK_STATUS_IGNORED):
        """Receiver loop - runs in a separate task"""
        with trio.CancelScope() as cscope:
            self.is_running = True
            task_status.started(cscope)
            try:
                while True:
                    msg = await self._conn.receive()
                    self._dispatch(msg)
            finally:
                self.is_running = False
                # Send errors to any tasks still waiting for a message.
                self._replies.drop_all()

                # Closing a memory channel can't block, but it only has an
                # async close method, so we need to shield it from cancellation.
                with trio.move_on_after(3) as cleanup_scope:
                    for filter in self._filters.filters.values():
                        cleanup_scope.shield = True
                        await filter.send_channel.aclose()


class Proxy(ProxyBase):
    """A trio proxy for calling D-Bus methods

    You can call methods on the proxy object, such as ``await bus_proxy.Hello()``
    to make a method call over D-Bus and wait for a reply. It will either
    return a tuple of returned data, or raise :exc:`.DBusErrorResponse`.
    The methods available are defined by the message generator you wrap.

    :param msggen: A message generator object.
    :param ~trio.DBusRouter router: Router to send and receive messages.
    """
    def __init__(self, msggen, router):
        super().__init__(msggen)
        if not isinstance(router, DBusRouter):
            raise TypeError("Proxy can only be used with DBusRequester")
        self._router = router

    def _method_call(self, make_msg):
        async def inner(*args, **kwargs):
            msg = make_msg(*args, **kwargs)
            assert msg.header.message_type is MessageType.method_call
            reply = await self._router.send_and_get_reply(msg)
            return unwrap_msg(reply)

        return inner


@asynccontextmanager
async def open_dbus_router(bus='SESSION', *, enable_fds=False):
    """Open a D-Bus 'router' to send and receive messages.

    Use as an async context manager::

        async with open_dbus_router() as req:
            ...

    :param str bus: 'SESSION' or 'SYSTEM' or a supported address.
    :return: :class:`DBusRouter`

    This is a shortcut for::

        conn = await open_dbus_connection()
        async with conn:
            async with conn.router() as req:
                ...
    """
    conn = await open_dbus_connection(bus, enable_fds=enable_fds)
    async with conn:
        async with conn.router() as rtr:
            yield rtr



import enum
import typing

from ._types import StrOrBytes
from .exceptions import ProtocolError, SOCKSError
from .utils import (
    AddressType,
    decode_address,
    encode_address,
    get_address_port_tuple_from_address,
)


class SOCKS4ReplyCode(bytes, enum.Enum):
    """Enumeration of SOCKS4 reply codes."""

    REQUEST_GRANTED = b"\x5A"
    REQUEST_REJECTED_OR_FAILED = b"\x5B"
    CONNECTION_FAILED = b"\x5C"
    AUTHENTICATION_FAILED = b"\x5D"


class SOCKS4Command(bytes, enum.Enum):
    """Enumeration of SOCKS4 command codes."""

    CONNECT = b"\x01"
    BIND = b"\x02"


class SOCKS4Request(typing.NamedTuple):
    """Encapsulates a request to the SOCKS4 proxy server

    Args:
        command: The command to request.
        port: The port number to connect to on the target host.
        addr: IP address of the target host.
        user_id: Optional user ID to be included in the request, if not supplied
            the user *must* provide one in the packing operation.
    """

    command: SOCKS4Command
    port: int
    addr: bytes
    user_id: typing.Optional[bytes] = None

    @classmethod
    def from_address(
        cls,
        command: SOCKS4Command,
        address: typing.Union[StrOrBytes, typing.Tuple[StrOrBytes, int]],
        user_id: typing.Optional[bytes] = None,
    ) -> "SOCKS4Request":
        """Convenience class method to build an instance from command and address.

        Args:
            command: The command to request.
            address: A string in the form 'HOST:PORT' or a tuple of ip address string
                and port number.
            user_id: Optional user ID.

        Returns:
            A SOCKS4Request instance.

        Raises:
            SOCKSError: If a domain name or IPv6 address was supplied.
        """
        address, port = get_address_port_tuple_from_address(address)
        atype, encoded_addr = encode_address(address)
        if atype != AddressType.IPV4:
            raise SOCKSError(
                "IPv6 addresses and domain names are not supported by SOCKS4"
            )
        return cls(command=command, addr=encoded_addr, port=port, user_id=user_id)

    def dumps(self, user_id: typing.Optional[bytes] = None) -> bytes:
        """Packs the instance into a raw binary in the appropriate form.

        Args:
            user_id: Optional user ID as an override, if not provided the instance's
                will be used, if none was provided at initialization an error is raised.

        Returns:
            The packed request.

        Raises:
            SOCKSError: If no user was specified in this call or on initialization.
        """
        user_id = user_id or self.user_id
        if user_id is None:
            raise SOCKSError("SOCKS4 requires a user_id, none was specified")

        return b"".join(
            [
                b"\x04",
                self.command,
                (self.port).to_bytes(2, byteorder="big"),
                self.addr,
                user_id,
                b"\x00",
            ]
        )


class SOCKS4ARequest(typing.NamedTuple):
    """Encapsulates a request to the SOCKS4A proxy server

    Args:
        command: The command to request.
        port: The port number to connect to on the target host.
        addr: IP address of the target host.
        user_id: Optional user ID to be included in the request, if not supplied
            the user *must* provide one in the packing operation.
    """

    command: SOCKS4Command
    port: int
    addr: bytes
    user_id: typing.Optional[bytes] = None

    @classmethod
    def from_address(
        cls,
        command: SOCKS4Command,
        address: typing.Union[StrOrBytes, typing.Tuple[StrOrBytes, int]],
        user_id: typing.Optional[bytes] = None,
    ) -> "SOCKS4ARequest":
        """Convenience class method to build an instance from command and address.

        Args:
            command: The command to request.
            address: A string in the form 'HOST:PORT' or a tuple of ip address string
                and port number.
            user_id: Optional user ID.

        Returns:
            A SOCKS4ARequest instance.
        """
        address, port = get_address_port_tuple_from_address(address)
        atype, encoded_addr = encode_address(address)
        return cls(command=command, addr=encoded_addr, port=port, user_id=user_id)

    def dumps(self, user_id: typing.Optional[bytes] = None) -> bytes:
        """Packs the instance into a raw binary in the appropriate form.

        Args:
            user_id: Optional user ID as an override, if not provided the instance's
                will be used, if none was provided at initialization an error is raised.

        Returns:
            The packed request.

        Raises:
            SOCKSError: If no user was specified in this call or on initialization.
        """
        user_id = user_id or self.user_id
        if user_id is None:
            raise SOCKSError("SOCKS4 requires a user_id, none was specified")

        return b"".join(
            [
                b"\x04",
                self.command,
                (self.port).to_bytes(2, byteorder="big"),
                b"\x00\x00\x00\xFF",  # arbitrary final non-zero byte
                user_id,
                b"\x00",
                self.addr,
                b"\x00",
            ]
        )


class SOCKS4Reply(typing.NamedTuple):
    """Encapsulates a reply from the SOCKS4 proxy server

    Args:
        reply_code: The code representing the type of reply.
        port: The port number returned.
        addr: Optional IP address returned.
    """

    reply_code: SOCKS4ReplyCode
    port: int
    addr: typing.Optional[str]

    @classmethod
    def loads(cls, data: bytes) -> "SOCKS4Reply":
        """Unpacks the reply data into an instance.

        Returns:
            The unpacked reply instance.

        Raises:
            ProtocolError: If the data does not match the spec.
        """
        if len(data) != 8 or data[0:1] != b"\x00":
            raise ProtocolError("Malformed reply")

        try:
            return cls(
                reply_code=SOCKS4ReplyCode(data[1:2]),
                port=int.from_bytes(data[2:4], byteorder="big"),
                addr=decode_address(AddressType.IPV4, data[4:8]),
            )
        except ValueError as exc:
            raise ProtocolError("Malformed reply") from exc


class SOCKS4Connection:
    """Encapsulates a SOCKS4 and SOCKS4A connection.

    Packs request objects into data suitable to be send and unpacks reply
    data into their appropriate reply objects.

    Args:
        user_id: The user ID to be sent as part of the requests.
    """

    def __init__(self, user_id: bytes):
        self.user_id = user_id

        self._data_to_send = bytearray()
        self._received_data = bytearray()

    def send(self, request: typing.Union[SOCKS4Request, SOCKS4ARequest]) -> None:
        """Packs a request object and adds it to the send data buffer.

        Args:
            request: The request instance to be packed.
        """
        user_id = request.user_id or self.user_id
        self._data_to_send += request.dumps(user_id=user_id)

    def receive_data(self, data: bytes) -> SOCKS4Reply:
        """Unpacks response data into a reply object.

        Args:
            data: The raw response data from the proxy server.

        Returns:
            The appropriate reply object.
        """
        self._received_data += data
        return SOCKS4Reply.loads(bytes(self._received_data))

    def data_to_send(self) -> bytes:
        """Returns the data to be sent via the I/O library of choice.

        Also clears the connection's buffer.
        """
        data = bytes(self._data_to_send)
        self._data_to_send = bytearray()
        return data



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help='Remove Python distributions')
@click.argument('names', required=True, nargs=-1)
@click.option('--dir', '-d', 'directory', help='The directory in which distributions reside')
@click.pass_obj
def remove(app: Application, *, names: tuple[str, ...], directory: str | None):
    """
    Remove Python distributions.

    You may select `all` to remove all installed distributions:

    \b
    ```
    hatch python remove all
    ```
    """
    manager = app.get_python_manager(directory)
    installed = manager.get_installed()
    selection = tuple(installed) if 'all' in names else names
    for name in selection:
        if name not in installed:
            app.display_warning(f'Distribution is not installed: {name}')
            continue

        dist = installed[name]
        with app.status(f'Removing {name}'):
            manager.remove(dist)



"""Code to communicate with PyPI to register distributions and upload files.

This is cribbed heavily from distutils.command.(upgrade|register), which as part
of Python is under the PSF license.
"""
import configparser
import getpass
import hashlib
import logging
import os
from pathlib import Path
import re
import requests
import sys
from dataclasses import dataclass
from typing import Optional
from urllib.parse import urlparse

from flit_core.common import make_metadata, Metadata, Module
from .config import read_flit_config

log = logging.getLogger(__name__)

PYPI = "https://upload.pypi.org/legacy/"
PYPIRC_DEFAULT = "~/.pypirc"

SWITCH_TO_HTTPS = (
    "http://pypi.python.org/",
    "http://testpypi.python.org/",
    "http://upload.pypi.org/",
)


@dataclass
class RepoDetails:
    url: str
    username: Optional[str] = None
    password: Optional[str] = None

    @property
    def is_pypi(self):
        # Main PyPI (pypi.org) or TestPyPI (test.pypi.org)
        # This is used to guess the URL for the project's HTML page
        return self.url.rstrip('/').endswith('/legacy')


def get_repositories(file="~/.pypirc"):
    """Get the known repositories from a pypirc file.

    This returns a dict keyed by name, of dicts with keys 'url', 'username',
    'password'. Username and password may be None.
    """
    cp = configparser.ConfigParser(interpolation=None)
    if isinstance(file, str):
        file = os.path.expanduser(file)

        if not os.path.isfile(file):
            return {'pypi': RepoDetails(url=PYPI)}

        cp.read(file)
    else:
        cp.read_file(file)

    names = cp.get('distutils', 'index-servers', fallback='pypi').split()

    repos = {}

    for name in names:
        repos[name] = RepoDetails(
            url=cp.get(name, 'repository', fallback=PYPI),
            username=cp.get(name, 'username', fallback=None),
            password=cp.get(name, 'password', fallback=None),
        )

    return repos


def get_repository(pypirc_path="~/.pypirc", name=None, project_name=None):
    """Get the url, username and password for one repository.

    Returns a dict with keys 'url', 'username', 'password'.

    There is a hierarchy of possible sources of information:

    Index URL:
    1. Command line arg --repository (looked up in .pypirc)
    2. $FLIT_INDEX_URL
    3. Repository called 'pypi' from .pypirc
    4. Default PyPI (hardcoded)

    Username:
    1. $FLIT_USERNAME
    2. Repository with specified name (default 'pypi') in .pypirc
    3. Terminal prompt (write to .pypirc if it doesn't exist yet)

    Password:
    1. $FLIT_PASSWORD
    2. Repository with specified name (default 'pypi') in .pypirc
    3. keyring - pypi_token:project:<project_name>
    4. keyring - pypi_token:user:<username>
    5. keyring - username
    6. Terminal prompt (store to keyring if available)
    """
    log.debug("Loading repositories config from %r", pypirc_path)
    repos_cfg = get_repositories(pypirc_path)

    if name is not None:
        repo = repos_cfg[name]
        if 'FLIT_INDEX_URL' in os.environ:
            raise EnvironmentError(
                "Use either FLIT_INDEX_URL or --repository, not both"
            )
    elif 'FLIT_INDEX_URL' in os.environ:
        repo = RepoDetails(url=os.environ['FLIT_INDEX_URL'])
    elif 'pypi' in repos_cfg:
        repo = repos_cfg['pypi']
    else:
        repo = RepoDetails(url=PYPI)

    if repo.url.startswith(SWITCH_TO_HTTPS):
        # Use https for PyPI, even if an http URL was given
        repo.url = 'https' + repo.url[4:]
    elif repo.url.startswith('http://'):
        log.warning("Unencrypted connection - credentials may be visible on "
                    "the network.")

    log.info("Using repository at %s", repo.url)

    if 'FLIT_USERNAME' in os.environ:
        repo.username = os.environ['FLIT_USERNAME']
    if not repo.username and sys.stdin.isatty():
        while not repo.username:
            repo.username = input("Username: ")
        if repo.url == PYPI:
            write_pypirc(repo, pypirc_path)
    elif not repo.username:
        raise Exception("Could not find username for upload.")

    if 'FLIT_PASSWORD' in os.environ:
        repo.password = os.environ['FLIT_PASSWORD']

    if not repo.password:
        token = find_token(repo, project_name)
        if token is not None:
            repo.username = '__token__'
            repo.password = token
        else:
            repo.password = get_password(repo)

    return repo

def write_pypirc(repo, file="~/.pypirc"):
    """Write .pypirc if it doesn't already exist
    """
    file = os.path.expanduser(file)
    if os.path.isfile(file):
        return

    with open(file, 'w', encoding='utf-8') as f:
        f.write("[pypi]\n"
                "username = %s\n" % repo.username)

def get_password(repo: RepoDetails):
    try:
        import keyring, keyring.errors
    except ImportError:  # pragma: no cover
        log.warning("Install keyring to store tokens/passwords securely")
        keyring = None
    else:
        try:
            stored_pw = keyring.get_password(repo.url, repo.username)
            if stored_pw is not None:
                return stored_pw
        except keyring.errors.KeyringError as e:
            log.warning("Could not get password from keyring (%s)", e)

    if sys.stdin.isatty():
        pw = None
        while not pw:
            print('Server  :', repo.url)
            print('Username:', repo.username)
            pw = getpass.getpass('Password: ')
    else:
        raise Exception("Could not find password for upload.")

    if keyring is not None:
        try:
            keyring.set_password(repo.url, repo.username, pw)
            log.info("Stored password with keyring")
        except keyring.errors.KeyringError as e:
            log.warning("Could not store password in keyring (%s)", e)

    return pw


def find_token(repo: RepoDetails, project_name: str):
    # https://packaging.python.org/en/latest/specifications/name-normalization/
    project_name = re.sub(r"[-_.]+", "-", project_name).lower()
    candidate_keys = [f"pypi_token:project:{project_name}"]

    if repo.username is not None:
        candidate_keys.append(f"pypi_token:user:{repo.username}")

    try:
        import keyring, keyring.errors
    except ImportError:  # pragma: no cover
        pass
    else:
        try:
            for key in candidate_keys:
                token = keyring.get_password(repo.url, key)
                if token is not None:
                    return token
        except keyring.errors.KeyringError as e:
            log.warning("Could not get token from keyring (%s)", e)


def build_post_data(action, metadata:Metadata):
    """Prepare the metadata needed for requests to PyPI.
    """
    d = {
        ":action": action,

        "name": metadata.name,
        "version": metadata.version,

        # additional meta-data
        "metadata_version": '2.3',
        "summary": metadata.summary,
        "home_page": metadata.home_page,
        "author": metadata.author,
        "author_email": metadata.author_email,
        "maintainer": metadata.maintainer,
        "maintainer_email": metadata.maintainer_email,
        "license": metadata.license,
        "description": metadata.description,
        "keywords": metadata.keywords,
        "platform": metadata.platform,
        "classifiers": metadata.classifiers,
        "download_url": metadata.download_url,
        "supported_platform": metadata.supported_platform,
        # Metadata 1.1 (PEP 314)
        "provides": metadata.provides,
        "requires": metadata.requires,
        "obsoletes": metadata.obsoletes,
        # Metadata 1.2 (PEP 345)
        "project_urls": metadata.project_urls,
        "provides_dist": metadata.provides_dist,
        "obsoletes_dist": metadata.obsoletes_dist,
        "requires_dist": metadata.requires_dist,
        "requires_external": metadata.requires_external,
        "requires_python": metadata.requires_python,
        # Metadata 2.1 (PEP 566)
        "description_content_type": metadata.description_content_type,
        "provides_extra": metadata.provides_extra,
      }

    return {k:v for k,v in d.items() if v}

def upload_file(file:Path, metadata:Metadata, repo: RepoDetails):
    """Upload a file to an index server, given the index server details.
    """
    data = build_post_data('file_upload', metadata)
    data['protocol_version'] = '1'
    if file.suffix == '.whl':
        data['filetype'] = 'bdist_wheel'
        py2_support = not (metadata.requires_python or '')\
                                    .startswith(('3', '>3', '>=3'))
        data['pyversion'] = ('py2.' if py2_support else '') + 'py3'
    else:
        data['filetype'] = 'sdist'

    with file.open('rb') as f:
        content = f.read()
        files = {'content': (file.name, content)}
        data['md5_digest'] = hashlib.md5(content).hexdigest()
        data['sha256_digest'] = hashlib.sha256(content).hexdigest()

    log.info('Uploading %s...', file)
    resp = requests.post(
        repo.url, data=data, files=files, auth=(repo.username, repo.password),
    )
    resp.raise_for_status()


def do_upload(file:Path, metadata:Metadata, repo: RepoDetails):
    """Upload a file to an index server.
    """
    upload_file(file, metadata, repo)

    if repo.is_pypi:
        domain = urlparse(repo.url).netloc
        if domain.startswith('upload.'):
            domain = domain[7:]
        log.info("Package is at https://%s/project/%s/", domain, metadata.name)
    else:
        log.info("Package is at %s/%s", repo.url, metadata.name)


def main(ini_path, repo_name, pypirc_path=None, formats=None, gen_setup_py=True,
         use_vcs=True):
    """Build and upload wheel and sdist."""
    if pypirc_path is None:
        pypirc_path = PYPIRC_DEFAULT
    elif not os.path.isfile(pypirc_path):
        raise FileNotFoundError("The specified pypirc config file does not exist.")

    ini_info = read_flit_config(ini_path)
    srcdir = ini_path.parent
    module = Module(ini_info.module, srcdir)
    metadata = make_metadata(module, ini_info)
    repo = get_repository(pypirc_path, repo_name, project_name=metadata.name)

    from . import build
    built = build.main(
        ini_path, formats=formats, gen_setup_py=gen_setup_py, use_vcs=use_vcs
    )

    if built.wheel is not None:
        do_upload(built.wheel.file, built.wheel.builder.metadata, repo)
    if built.sdist is not None:
        do_upload(built.sdist.file, built.sdist.builder.metadata, repo)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help='Build a project')
@click.argument('location', required=False)
@click.option(
    '--target',
    '-t',
    'targets',
    multiple=True,
    help=(
        'The target to build, overriding project defaults. '
        'This may be selected multiple times e.g. `-t sdist -t wheel`'
    ),
)
@click.option(
    '--hooks-only', is_flag=True, help='Whether or not to only execute build hooks [env var: `HATCH_BUILD_HOOKS_ONLY`]'
)
@click.option(
    '--no-hooks', is_flag=True, help='Whether or not to disable build hooks [env var: `HATCH_BUILD_NO_HOOKS`]'
)
@click.option(
    '--ext',
    is_flag=True,
    help=(
        'Whether or not to only execute build hooks for distributing binary Python packages, such as '
        'compiling extensions. Equivalent to `--hooks-only -t wheel`'
    ),
)
@click.option(
    '--clean',
    '-c',
    is_flag=True,
    help='Whether or not existing artifacts should first be removed [env var: `HATCH_BUILD_CLEAN`]',
)
@click.option(
    '--clean-hooks-after',
    is_flag=True,
    help=(
        'Whether or not build hook artifacts should be removed after each build '
        '[env var: `HATCH_BUILD_CLEAN_HOOKS_AFTER`]'
    ),
)
@click.option('--clean-only', is_flag=True, hidden=True)
@click.pass_obj
def build(app: Application, location, targets, hooks_only, no_hooks, ext, clean, clean_hooks_after, clean_only):
    """Build a project."""
    app.ensure_environment_plugin_dependencies()

    from hatch.config.constants import AppEnvVars
    from hatch.project.config import env_var_enabled
    from hatch.project.constants import BUILD_BACKEND, DEFAULT_BUILD_DIRECTORY, BuildEnvVars
    from hatch.utils.fs import Path
    from hatch.utils.runner import ExecutionContext
    from hatch.utils.structures import EnvVars

    build_dir = Path(location).resolve() if location else None
    if ext:
        hooks_only = True
        targets = ('wheel',)
    elif not targets:
        targets = ('sdist', 'wheel')

    env_vars = {}
    if app.verbose:
        env_vars[AppEnvVars.VERBOSE] = str(app.verbosity)
    elif app.quiet:
        env_vars[AppEnvVars.QUIET] = str(abs(app.verbosity))

    with EnvVars(env_vars):
        app.project.prepare_build_environment(targets=[target.split(':')[0] for target in targets])

    build_backend = app.project.metadata.build.build_backend
    with app.project.location.as_cwd(), app.project.build_env.get_env_vars():
        for target in targets:
            target_name, _, _ = target.partition(':')
            if not clean_only:
                app.display_header(target_name)

            if build_backend != BUILD_BACKEND:
                if target_name == 'sdist':
                    directory = build_dir or app.project.location / DEFAULT_BUILD_DIRECTORY
                    directory.ensure_dir_exists()
                    artifact_path = app.project.build_frontend.build_sdist(directory)
                elif target_name == 'wheel':
                    directory = build_dir or app.project.location / DEFAULT_BUILD_DIRECTORY
                    directory.ensure_dir_exists()
                    artifact_path = app.project.build_frontend.build_wheel(directory)
                else:
                    app.abort(f'Target `{target_name}` is not supported by `{build_backend}`')

                app.display_info(
                    str(artifact_path.relative_to(app.project.location))
                    if app.project.location in artifact_path.parents
                    else str(artifact_path)
                )
            else:
                command = ['python', '-u', '-m', 'hatchling', 'build', '--target', target]

                # We deliberately pass the location unchanged so that absolute paths may be non-local
                # and reflect wherever builds actually take place
                if location:
                    command.extend(('--directory', location))

                if hooks_only or env_var_enabled(BuildEnvVars.HOOKS_ONLY):
                    command.append('--hooks-only')

                if no_hooks or env_var_enabled(BuildEnvVars.NO_HOOKS):
                    command.append('--no-hooks')

                if clean or env_var_enabled(BuildEnvVars.CLEAN):
                    command.append('--clean')

                if clean_hooks_after or env_var_enabled(BuildEnvVars.CLEAN_HOOKS_AFTER):
                    command.append('--clean-hooks-after')

                if clean_only:
                    command.append('--clean-only')

                context = ExecutionContext(app.project.build_env)
                context.add_shell_command(command)
                context.env_vars.update(env_vars)
                app.execute_context(context)



from __future__ import annotations

import argparse

from pdm.cli.commands.base import BaseCommand
from pdm.cli.commands.venv.activate import ActivateCommand
from pdm.cli.commands.venv.create import CreateCommand
from pdm.cli.commands.venv.list import ListCommand
from pdm.cli.commands.venv.purge import PurgeCommand
from pdm.cli.commands.venv.remove import RemoveCommand
from pdm.cli.commands.venv.utils import get_venv_with_name
from pdm.cli.options import project_option
from pdm.project import Project


class Command(BaseCommand):
    """Virtualenv management"""

    name = "venv"
    arguments = (project_option,)

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        group = parser.add_mutually_exclusive_group()
        group.add_argument("--path", help="Show the path to the given virtualenv")
        group.add_argument("--python", help="Show the python interpreter path for the given virtualenv")
        subparser = parser.add_subparsers(title="commands", metavar="")
        CreateCommand.register_to(subparser, "create")
        ListCommand.register_to(subparser, "list")
        RemoveCommand.register_to(subparser, "remove")
        ActivateCommand.register_to(subparser, "activate")
        PurgeCommand.register_to(subparser, "purge")
        self.parser = parser

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        if options.path:
            venv = get_venv_with_name(project, options.path)
            project.core.ui.echo(str(venv.root))
        elif options.python:
            venv = get_venv_with_name(project, options.python)
            project.core.ui.echo(str(venv.interpreter))
        else:
            self.parser.print_help()



import click


@click.group(short_help='Manage environment dependencies')
def dep():
    pass


@dep.command('hash', short_help='Output a hash of the currently defined dependencies')
@click.option('--project-only', '-p', is_flag=True, help='Whether or not to exclude environment dependencies')
@click.option('--env-only', '-e', is_flag=True, help='Whether or not to exclude project dependencies')
@click.pass_obj
def hash_dependencies(app, project_only, env_only):
    """Output a hash of the currently defined dependencies."""
    app.ensure_environment_plugin_dependencies()

    from hatch.utils.dep import get_complex_dependencies, hash_dependencies

    environment = app.project.get_environment()

    all_requirements = []
    if project_only:
        dependencies, _ = app.project.get_dependencies()
        dependencies_complex = get_complex_dependencies(dependencies)
        all_requirements.extend(dependencies_complex.values())
    elif env_only:
        all_requirements.extend(environment.environment_dependencies_complex)
    else:
        dependencies, _ = app.project.get_dependencies()
        dependencies_complex = get_complex_dependencies(dependencies)
        all_requirements.extend(dependencies_complex.values())
        all_requirements.extend(environment.environment_dependencies_complex)

    app.display(hash_dependencies(all_requirements))


@dep.group(short_help='Display dependencies in various formats')
def show():
    pass


@show.command(short_help='Enumerate dependencies in a tabular format')
@click.option('--project-only', '-p', is_flag=True, help='Whether or not to exclude environment dependencies')
@click.option('--env-only', '-e', is_flag=True, help='Whether or not to exclude project dependencies')
@click.option('--lines', '-l', 'show_lines', is_flag=True, help='Whether or not to show lines between table rows')
@click.option('--ascii', 'force_ascii', is_flag=True, help='Whether or not to only use ASCII characters')
@click.pass_obj
def table(app, project_only, env_only, show_lines, force_ascii):
    """Enumerate dependencies in a tabular format."""
    app.ensure_environment_plugin_dependencies()

    from packaging.requirements import Requirement

    from hatch.utils.dep import get_complex_dependencies, get_normalized_dependencies, normalize_marker_quoting

    environment = app.project.get_environment()

    project_requirements = []
    environment_requirements = []
    if project_only:
        dependencies, _ = app.project.get_dependencies()
        dependencies_complex = get_complex_dependencies(dependencies)
        project_requirements.extend(dependencies_complex.values())
    elif env_only:
        environment_requirements.extend(environment.environment_dependencies_complex)
    else:
        dependencies, _ = app.project.get_dependencies()
        dependencies_complex = get_complex_dependencies(dependencies)
        project_requirements.extend(dependencies_complex.values())
        environment_requirements.extend(environment.environment_dependencies_complex)

    for all_requirements, table_title in (
        (project_requirements, 'Project'),
        (environment_requirements, f'Env: {app.env}'),
    ):
        if not all_requirements:
            continue

        normalized_requirements = [Requirement(d) for d in get_normalized_dependencies(all_requirements)]

        columns = {'Name': {}, 'URL': {}, 'Versions': {}, 'Markers': {}, 'Features': {}}
        for i, requirement in enumerate(normalized_requirements):
            columns['Name'][i] = requirement.name

            if requirement.url:
                columns['URL'][i] = str(requirement.url)

            if requirement.specifier:
                columns['Versions'][i] = str(requirement.specifier)

            if requirement.marker:
                columns['Markers'][i] = normalize_marker_quoting(str(requirement.marker))

            if requirement.extras:
                columns['Features'][i] = ', '.join(sorted(requirement.extras))

        column_options = {}
        for column_title in columns:
            if column_title != 'URL':
                column_options[column_title] = {'no_wrap': True}

        app.display_table(
            table_title, columns, show_lines=show_lines, column_options=column_options, force_ascii=force_ascii
        )


@show.command(short_help='Enumerate dependencies as a list of requirements')
@click.option('--project-only', '-p', is_flag=True, help='Whether or not to exclude environment dependencies')
@click.option('--env-only', '-e', is_flag=True, help='Whether or not to exclude project dependencies')
@click.option(
    '--feature',
    '-f',
    'features',
    multiple=True,
    help='Whether or not to only show the dependencies of the specified features',
)
@click.option('--all', 'all_features', is_flag=True, help='Whether or not to include the dependencies of all features')
@click.pass_obj
def requirements(app, project_only, env_only, features, all_features):
    """Enumerate dependencies as a list of requirements."""
    app.ensure_environment_plugin_dependencies()

    from hatch.utils.dep import get_complex_dependencies, get_complex_features, get_normalized_dependencies
    from hatchling.metadata.utils import normalize_project_name

    environment = app.project.get_environment()
    dependencies, optional_dependencies = app.project.get_dependencies()
    dependencies_complex = get_complex_dependencies(dependencies)
    optional_dependencies_complex = get_complex_features(optional_dependencies)

    all_requirements = []
    if features:
        for raw_feature in features:
            feature = normalize_project_name(raw_feature)
            if feature not in optional_dependencies_complex:
                app.abort(f'Feature `{feature}` is not defined in field `project.optional-dependencies`')

            all_requirements.extend(optional_dependencies_complex[feature].values())
    elif project_only:
        all_requirements.extend(dependencies_complex.values())
    elif env_only:
        all_requirements.extend(environment.environment_dependencies_complex)
    else:
        all_requirements.extend(dependencies_complex.values())
        all_requirements.extend(environment.environment_dependencies_complex)

    if not features and all_features:
        for optional_dependencies in optional_dependencies_complex.values():
            all_requirements.extend(optional_dependencies.values())

    for dependency in get_normalized_dependencies(all_requirements):
        app.display(dependency)



from __future__ import annotations

import os
import platform
import sys
from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING

from hatch.config.constants import PythonEnvVars
from hatch.errors import PythonDistributionResolutionError, PythonDistributionUnknownError
from hatch.python.distributions import DISTRIBUTIONS, ORDERED_DISTRIBUTIONS

if TYPE_CHECKING:
    from packaging.version import Version

    from hatch.utils.fs import Path


# Use an artificially high epoch to ensure that custom distributions are always considered newer
CUSTOM_DISTRIBUTION_VERSION_EPOCH = 100


def custom_env_var(prefix: str, name: str) -> str:
    return f'{prefix}{name.upper().replace(".", "_")}'


def get_custom_source(name: str) -> str | None:
    return os.environ.get(custom_env_var(PythonEnvVars.CUSTOM_SOURCE_PREFIX, name))


def get_custom_version(name: str) -> str | None:
    return os.environ.get(custom_env_var(PythonEnvVars.CUSTOM_VERSION_PREFIX, name))


def get_custom_path(name: str) -> str | None:
    return os.environ.get(custom_env_var(PythonEnvVars.CUSTOM_PATH_PREFIX, name))


class Distribution(ABC):
    def __init__(self, name: str, source: str) -> None:
        self.__name = name
        self.__source = source

    @property
    def name(self) -> str:
        return self.__name

    @cached_property
    def source(self) -> str:
        return self.__source if (custom_source := get_custom_source(self.name)) is None else custom_source

    @cached_property
    def archive_name(self) -> str:
        return self.source.rsplit('/', 1)[-1]

    def unpack(self, archive: Path, directory: Path) -> None:
        if self.source.endswith('.zip'):
            import zipfile

            with zipfile.ZipFile(archive, 'r') as zf:
                zf.extractall(directory)
        elif self.source.endswith(('.tar.gz', '.tgz')):
            import tarfile

            with tarfile.open(archive, 'r:gz') as tf:
                if sys.version_info[:2] >= (3, 12):
                    tf.extractall(directory, filter='data')
                else:
                    tf.extractall(directory)  # noqa: S202
        elif self.source.endswith(('.tar.bz2', '.bz2')):
            import tarfile

            with tarfile.open(archive, 'r:bz2') as tf:
                if sys.version_info[:2] >= (3, 12):
                    tf.extractall(directory, filter='data')
                else:
                    tf.extractall(directory)  # noqa: S202
        elif self.source.endswith(('.tar.zst', '.tar.zstd')):
            import tarfile

            import zstandard

            with open(archive, 'rb') as ifh:
                dctx = zstandard.ZstdDecompressor()
                with dctx.stream_reader(ifh) as reader, tarfile.open(mode='r|', fileobj=reader) as tf:
                    if sys.version_info[:2] >= (3, 12):
                        tf.extractall(directory, filter='data')
                    else:
                        tf.extractall(directory)  # noqa: S202
        else:
            message = f'Unknown archive type: {archive}'
            raise ValueError(message)

    @property
    @abstractmethod
    def version(self) -> Version:
        pass

    @property
    @abstractmethod
    def python_path(self) -> str:
        pass


class CPythonStandaloneDistribution(Distribution):
    @cached_property
    def version(self) -> Version:
        from packaging.version import Version

        if (custom_version := get_custom_version(self.name)) is not None:
            return Version(f'{CUSTOM_DISTRIBUTION_VERSION_EPOCH}!{custom_version}')

        # .../cpython-3.12.0%2B20231002-...
        # .../cpython-3.7.9-...
        _, _, remaining = self.source.partition('/cpython-')
        # 3.12.0%2B20231002-...
        # 3.7.9-...
        version = remaining.split('%2B')[0] if '%2B' in remaining else remaining.split('-')[0]
        return Version(f'0!{version}')

    @cached_property
    def python_path(self) -> str:
        if (custom_path := get_custom_path(self.name)) is not None:
            return custom_path

        if self.name == '3.7':
            if sys.platform == 'win32':
                return r'python\install\python.exe'

            return 'python/install/bin/python3'

        if sys.platform == 'win32':
            return r'python\python.exe'

        return 'python/bin/python3'


class PyPyOfficialDistribution(Distribution):
    @cached_property
    def version(self) -> Version:
        from packaging.version import Version

        if (custom_version := get_custom_version(self.name)) is not None:
            return Version(f'{CUSTOM_DISTRIBUTION_VERSION_EPOCH}!{custom_version}')

        *_, remaining = self.source.partition('/pypy/')
        _, version, *_ = remaining.split('-')
        return Version(f'0!{version[1:]}')

    @cached_property
    def python_path(self) -> str:
        if (custom_path := get_custom_path(self.name)) is not None:
            return custom_path

        directory = self.archive_name
        for extension in ('.tar.bz2', '.zip'):
            if directory.endswith(extension):
                directory = directory[: -len(extension)]
                break

        if sys.platform == 'win32':
            return rf'{directory}\pypy.exe'

        return f'{directory}/bin/pypy'


def get_distribution(name: str, source: str = '', variant_cpu: str = '', variant_gil: str = '') -> Distribution:
    if source:
        return _get_distribution_class(source)(name, source)

    if name not in DISTRIBUTIONS:
        message = f'Unknown distribution: {name}'
        raise PythonDistributionUnknownError(message)

    arch = platform.machine().lower()
    if sys.platform == 'win32':
        system = 'windows'
        abi = 'msvc'
    elif sys.platform == 'darwin':
        system = 'macos'
        abi = ''
    else:
        system = 'linux'
        abi = 'gnu' if any(platform.libc_ver()) else 'musl'

    if not variant_cpu:
        variant_cpu = _get_default_variant_cpu(name, system, arch)

    if not variant_gil:
        variant_gil = _get_default_variant_gil()

    key = (system, arch, abi, variant_cpu, variant_gil)

    keys: dict[tuple, str] = DISTRIBUTIONS[name]
    if key not in keys:
        message = f'Could not find a default source for {name=} {system=} {arch=} {abi=} {variant_cpu=} {variant_gil=}'
        raise PythonDistributionResolutionError(message)

    source = keys[key]
    return _get_distribution_class(source)(name, source)


def get_compatible_distributions() -> dict[str, Distribution]:
    distributions: dict[str, Distribution] = {}
    for name in ORDERED_DISTRIBUTIONS:
        try:
            dist = get_distribution(name)
        except PythonDistributionResolutionError:
            pass
        else:
            distributions[name] = dist

    return distributions


def _guess_linux_variant_cpu() -> str:
    # Use the highest that we know is most common when we can't parse CPU data
    default = 'v3'
    try:
        # Don't use our utility Path so we can properly mock
        with open('/proc/cpuinfo', encoding='utf-8') as f:
            contents = f.read()
    except OSError:
        return default

    # See https://clang.llvm.org/docs/UsersManual.html#x86 for the
    # instructions for each architecture variant and
    # https://github.com/torvalds/linux/blob/master/arch/x86/include/asm/cpufeatures.h
    # for the corresponding Linux flags
    v2_flags = {'cx16', 'lahf_lm', 'popcnt', 'pni', 'sse4_1', 'sse4_2', 'ssse3'}
    v3_flags = {'avx', 'avx2', 'bmi1', 'bmi2', 'f16c', 'fma', 'movbe', 'xsave'} | v2_flags
    v4_flags = {'avx512f', 'avx512bw', 'avx512cd', 'avx512dq', 'avx512vl'} | v3_flags

    for line in contents.splitlines():
        key, _, value = line.partition(':')
        if key.strip() == 'flags':
            flags = set(value.strip().split())

            if flags.issuperset(v4_flags):
                return 'v4'

            if flags.issuperset(v3_flags):
                return 'v3'

            if flags.issuperset(v2_flags):
                return 'v2'

            return 'v1'

    return default


def _get_default_variant_cpu(name: str, system: str, arch: str) -> str:
    # not PyPy
    if name[0].isdigit():
        variant = os.environ.get(
            'HATCH_PYTHON_VARIANT_CPU',
            # Legacy name
            os.environ.get(f'HATCH_PYTHON_VARIANT_{system.upper()}', ''),
        ).lower()

        # https://gregoryszorc.com/docs/python-build-standalone/main/running.html
        if system == 'linux' and arch == 'x86_64':
            # Intel-specific optimizations depending on age of release
            if variant:
                return variant

            if name == '3.8':
                return 'v1'

            if name != '3.7':
                return _guess_linux_variant_cpu()

    return ''


def _get_default_variant_gil() -> str:
    return os.environ.get('HATCH_PYTHON_VARIANT_GIL', '').lower()


def _get_distribution_class(source: str) -> type[Distribution]:
    if source.startswith('https://github.com/indygreg/python-build-standalone/releases/download/'):
        return CPythonStandaloneDistribution

    if source.startswith('https://downloads.python.org/pypy/'):
        return PyPyOfficialDistribution

    message = f'Unknown distribution source: {source}'
    raise ValueError(message)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


def ensure_path_public(path: str, shells: list[str]) -> bool:
    import userpath

    if userpath.in_current_path(path) or userpath.in_new_path(path, shells):
        return True

    userpath.append(path, shells=shells)
    return False


@click.command(short_help='Install Python distributions')
@click.argument('names', required=True, nargs=-1)
@click.option('--private', is_flag=True, help='Do not add distributions to the user PATH')
@click.option('--update', '-u', is_flag=True, help='Update existing installations')
@click.option(
    '--dir', '-d', 'directory', help='The directory in which to install distributions, overriding configuration'
)
@click.pass_obj
def install(app: Application, *, names: tuple[str, ...], private: bool, update: bool, directory: str | None):
    """
    Install Python distributions.

    You may select `all` to install all compatible distributions:

    \b
    ```
    hatch python install all
    ```

    You can set custom sources for distributions by setting the `HATCH_PYTHON_SOURCE_<NAME>` environment variable
    where `<NAME>` is the uppercased version of the distribution name with periods replaced by underscores e.g.
    `HATCH_PYTHON_SOURCE_PYPY3_10`.
    """
    from hatch.errors import PythonDistributionResolutionError, PythonDistributionUnknownError
    from hatch.python.distributions import ORDERED_DISTRIBUTIONS
    from hatch.python.resolve import get_distribution

    shells = []
    if not private and not app.platform.windows:
        shell_name, _ = app.shell_data
        shells.append(shell_name)

    manager = app.get_python_manager(directory)
    installed = manager.get_installed()
    selection = ORDERED_DISTRIBUTIONS if 'all' in names else names
    unknown = []
    compatible = []
    incompatible = []
    for name in selection:
        if name in installed:
            compatible.append(name)
            continue

        try:
            get_distribution(name)
        except PythonDistributionUnknownError:
            unknown.append(name)
        except PythonDistributionResolutionError:
            incompatible.append(name)
        else:
            compatible.append(name)

    if unknown:
        app.abort(f'Unknown distributions: {", ".join(unknown)}')
    elif incompatible and (not compatible or 'all' not in names):
        app.abort(f'Incompatible distributions: {", ".join(incompatible)}')

    directories_made_public = []
    for name in compatible:
        needs_update = False
        if name in installed:
            installed_dist = installed[name]
            needs_update = installed_dist.needs_update()
            if not needs_update:
                app.display_warning(f'The latest version is already installed: {installed_dist.version}')
                continue

            if not (update or app.confirm(f'Update {name}?')):
                app.abort(f'Distribution is already installed: {name}')

        with app.status(f'{"Updating" if needs_update else "Installing"} {name}'):
            dist = manager.install(name)
            if not private:
                python_directory = str(dist.python_path.parent)
                if not ensure_path_public(python_directory, shells=shells):
                    directories_made_public.append(python_directory)

        app.display_success(f'{"Updated" if needs_update else "Installed"} {name} @ {dist.path}')

    if directories_made_public:
        multiple = len(directories_made_public) > 1
        app.display(
            f'\nThe following director{"ies" if multiple else "y"} ha{"ve" if multiple else "s"} '
            f'been added to your PATH (pending a shell restart):\n'
        )
        for public_directory in directories_made_public:
            app.display(public_directory)



"""Sans-I/O implementation of SOCKS4, SOCKS4A, and SOCKS5."""

from .exceptions import ProtocolError, SOCKSError
from .socks4 import (
    SOCKS4ARequest,
    SOCKS4Command,
    SOCKS4Connection,
    SOCKS4Reply,
    SOCKS4ReplyCode,
    SOCKS4Request,
)
from .socks5 import (
    SOCKS5AType,
    SOCKS5AuthMethod,
    SOCKS5AuthMethodsRequest,
    SOCKS5AuthReply,
    SOCKS5Command,
    SOCKS5CommandRequest,
    SOCKS5Connection,
    SOCKS5Reply,
    SOCKS5ReplyCode,
    SOCKS5UsernamePasswordRequest,
)

__version__ = "1.0.0"

__all__ = [
    "SOCKS4Request",
    "SOCKS4ARequest",
    "SOCKS4Reply",
    "SOCKS4Connection",
    "SOCKS4Command",
    "SOCKS4ReplyCode",
    "SOCKS5AType",
    "SOCKS5AuthMethodsRequest",
    "SOCKS5AuthReply",
    "SOCKS5AuthMethod",
    "SOCKS5Connection",
    "SOCKS5Command",
    "SOCKS5CommandRequest",
    "SOCKS5ReplyCode",
    "SOCKS5Reply",
    "SOCKS5UsernamePasswordRequest",
    "SOCKSError",
    "ProtocolError",
]



from __future__ import annotations

import base64
import json
from typing import Any, cast

import hishel
from hishel._serializers import Metadata
from httpcore import Request, Response


class Encoder(json.JSONEncoder):
    """Expand standard json encoder to support dumps bytes object."""

    bytes_ident = "PDM_BYTES_OBJECT"

    def default(self, o: Any) -> Any:
        if isinstance(o, bytes):
            base64_string = base64.b64encode(o).decode()
            return {"type": self.bytes_ident, "val": base64_string}
        return super().default(o)

    @classmethod
    def object_hook(cls, obj: Any) -> Any:
        if isinstance(obj, dict) and obj.get("type") == cls.bytes_ident:
            val = obj.get("val")
            if val is not None and isinstance(val, str):
                return base64.b64decode(val)
        return obj


class JSONMsgPack:
    UnpackValueError = json.JSONDecodeError

    @staticmethod
    def packb(data: dict, use_bin_type: bool = True) -> bytes:
        return json.dumps(data, cls=Encoder).encode()

    @staticmethod
    def loads(data: bytes, raw: bool = False) -> Any:
        return json.loads(data, object_hook=Encoder.object_hook)


def _get_msgpack_implementation() -> type[JSONMsgPack]:
    try:
        import msgpack
    except ImportError:
        return JSONMsgPack
    else:
        return cast("type[JSONMsgPack]", msgpack)


class MsgPackSerializer(hishel.BaseSerializer):
    KNOWN_REQUEST_EXTENSIONS = ("timeout", "sni_hostname")
    KNOWN_RESPONSE_EXTENSIONS = ("http_version", "reason_phrase")
    DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

    implementation = _get_msgpack_implementation()

    def dumps(self, response: Response, request: Request, metadata: Metadata) -> bytes:
        from hishel._utils import normalized_url

        response_dict = {
            "status": response.status,
            "headers": response.headers,
            "content": response.content,
            "extensions": {
                key: value for key, value in response.extensions.items() if key in self.KNOWN_RESPONSE_EXTENSIONS
            },
        }

        request_dict = {
            "method": request.method.decode("ascii"),
            "url": normalized_url(request.url),
            "headers": request.headers,
            "extensions": {
                key: value for key, value in request.extensions.items() if key in self.KNOWN_REQUEST_EXTENSIONS
            },
        }

        metadata_dict = {
            "cache_key": metadata["cache_key"],
            "number_of_uses": metadata["number_of_uses"],
            "created_at": metadata["created_at"].strftime(self.DATETIME_FORMAT),
        }

        full_dict = {
            "response": response_dict,
            "request": request_dict,
            "metadata": metadata_dict,
        }
        return cast(bytes, self.implementation.packb(full_dict, use_bin_type=True))

    def loads(self, data: bytes) -> tuple[Response, Request, Metadata] | None:
        from datetime import datetime

        try:
            full_dict = cast("dict[str, Any]", self.implementation.loads(data, raw=False))
        except UnicodeDecodeError:
            # For compatibility: loaded by json, while data was dumped by MsgPack
            return None
        except self.implementation.UnpackValueError:
            if not data.strip().startswith(b"{"):
                return None
            # Dumped by json, but tried to load by MsgPack
            try:
                full_dict = cast("dict[str, Any]", json.loads(data, object_hook=Encoder.object_hook))
            except json.JSONDecodeError:
                return None

        response_dict = full_dict["response"]
        request_dict = full_dict["request"]
        metadata_dict = full_dict["metadata"]
        metadata_dict["created_at"] = datetime.strptime(metadata_dict["created_at"], self.DATETIME_FORMAT)

        response = Response(
            status=response_dict["status"],
            headers=response_dict["headers"],
            content=response_dict["content"],
            extensions=response_dict["extensions"],
        )

        request = Request(
            method=request_dict["method"],
            url=request_dict["url"],
            headers=request_dict["headers"],
            extensions=request_dict["extensions"],
        )

        metadata = Metadata(
            cache_key=metadata_dict["cache_key"],
            created_at=metadata_dict["created_at"],
            number_of_uses=metadata_dict["number_of_uses"],
        )

        return response, request, metadata

    @property
    def is_binary(self) -> bool:
        return True



from __future__ import annotations

import os


def locate_file(root: str, file_name: str, *, boundary: str | None = None) -> str | None:
    while True:
        file_path = os.path.join(root, file_name)
        if os.path.isfile(file_path):
            return file_path

        if boundary is not None and os.path.exists(os.path.join(root, boundary)):
            return None

        new_root = os.path.dirname(root)
        if new_root == root:
            return None

        root = new_root


def path_to_uri(path: str) -> str:
    if os.sep == "/":
        return f"file://{os.path.abspath(path).replace(' ', '%20')}"

    return f"file:///{os.path.abspath(path).replace(' ', '%20').replace(os.sep, '/')}"



from __future__ import annotations

from typing import Any


def get_default_config() -> dict[str, Any]:
    return {
        'skip-install': True,
        'builder': True,
        'installer': 'uv',
    }



import pytest

from jeepney.wrappers import *

def test_bad_bus_name():
    obj = '/com/example/foo'
    DBusAddress(obj, 'com.example.a')  # Valid (well known name)
    DBusAddress(obj, 'com.example.a-b')  # Valid but discouraged
    DBusAddress(obj, ':1.13')  # Valid (unique name)

    with pytest.raises(ValueError, match='too long'):
        DBusAddress(obj, 'com.example.' + ('a' * 256))

    with pytest.raises(ValueError):
        DBusAddress(obj, '.com.example.a')

    with pytest.raises(ValueError):
        DBusAddress(obj, 'com..example.a')

    with pytest.raises(ValueError):
        DBusAddress(obj, 'com.2example.a')

    with pytest.raises(ValueError):
        DBusAddress(obj, 'cöm.example.a')  # Non-ASCII character

    with pytest.raises(ValueError):
        DBusAddress(obj, 'com')

def test_bad_interface():
    obj = '/com/example/foo'
    busname = 'com.example.foo'
    DBusAddress(obj, 'com.example.a', 'com.example.a_b')  # Valid

    with pytest.raises(ValueError, match='too long'):
        DBusAddress(obj, 'com.example.a', 'com.example.' + ('a' * 256))

    with pytest.raises(ValueError):
        DBusAddress(obj, 'com.example.a', 'com.example.a-b')  # No hyphens

    with pytest.raises(ValueError):
        DBusAddress(obj, busname, '.com.example.a')

    with pytest.raises(ValueError):
        DBusAddress(obj, busname, 'com..example.a')

    with pytest.raises(ValueError):
        DBusAddress(obj, busname, 'com.2example.a')

    with pytest.raises(ValueError):
        DBusAddress(obj, busname, 'cöm.example.a')  # Non-ASCII character

    with pytest.raises(ValueError):
        DBusAddress(obj, busname, 'com')


def test_bad_member_name():
    addr = DBusAddress(
        '/org/freedesktop/DBus',
        bus_name='org.freedesktop.DBus',
        interface='org.freedesktop.DBus',
    )
    new_method_call(addr, 'Hello')

    with pytest.raises(ValueError, match='too long'):
        new_method_call(addr, 'Hell' + ('o' * 256))

    with pytest.raises(ValueError):
        new_method_call(addr, 'org.Hello')

    with pytest.raises(ValueError):
        new_method_call(addr, '9Hello')

    with pytest.raises(ValueError):
        new_method_call(addr, '')



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(
    short_help='Install the latest version', context_settings={'help_option_names': [], 'ignore_unknown_options': True}
)
@click.argument('args', nargs=-1)
@click.pass_obj
def update(app: Application, *, args: tuple[str, ...]):  # noqa: ARG001
    app.abort('Hatch is not installed as a binary')



from hatch.template import File, files_default, find_template_files
from hatch.template.plugin.interface import TemplateInterface
from hatch.utils.fs import Path
from hatch.utils.network import download_file


class DefaultTemplate(TemplateInterface):
    PLUGIN_NAME = 'default'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.plugin_config.setdefault('ci', False)
        self.plugin_config.setdefault('src-layout', True)
        self.plugin_config.setdefault('tests', True)

    def initialize_config(self, config):
        # Default values
        config['readme_file_path'] = 'README.md'
        config['package_metadata_file_path'] = f'src/{config["package_name"]}/__about__.py'

        license_data = {}

        # Licenses
        license_ids = config['licenses']['default']
        if not license_ids:
            config['license_data'] = license_data
            config['license_expression'] = ''
            config['license_files'] = ''
            config['license_header'] = ''
            return

        cached_licenses_dir = self.cache_dir / 'licenses'
        cached_licenses_dir.ensure_dir_exists()

        license_ids = sorted(set(license_ids))
        for license_id in sorted(set(license_ids)):
            license_file_name = f'{license_id}.txt'
            cached_license_path = cached_licenses_dir / license_file_name
            if not cached_license_path.is_file():
                from packaging.licenses._spdx import VERSION  # noqa: PLC2701

                url = f'https://raw.githubusercontent.com/spdx/license-list-data/v{VERSION}/text/{license_file_name}'
                for _ in range(5):
                    try:
                        download_file(cached_license_path, url)
                    except Exception:  # noqa: BLE001, S112
                        continue
                    else:
                        break

            license_data[license_id] = cached_license_path.read_text(encoding='utf-8')

        config['license_data'] = license_data
        config['license_expression'] = ' OR '.join(license_data)
        config['license_header'] = (
            ''
            if not config['licenses']['headers']
            else f"""\
# SPDX-FileCopyrightText: {self.creation_time.year}-present {config['name']} <{config['email']}>
#
# SPDX-License-Identifier: {config['license_expression']}
"""
        )
        if len(license_ids) == 1:
            config['license_files'] = ''
        else:
            config['license_files'] = '\nlicense-files = { globs = ["LICENSES/*"] }'

        if config['args']['cli']:
            config['dependencies'].add('click')

        if not self.plugin_config['src-layout']:
            config['package_metadata_file_path'] = f'{config["package_metadata_file_path"][4:]}'

    def get_files(self, config):
        files = list(find_template_files(files_default))

        # Add any licenses
        license_data = config['license_data']
        if license_data:
            if len(license_data) == 1:
                license_id, text = next(iter(license_data.items()))
                license_text = get_license_text(config, license_id, text, self.creation_time)
                files.append(File(Path('LICENSE.txt'), license_text))
            else:
                # https://reuse.software/faq/#multi-licensing
                for license_id, text in license_data.items():
                    license_text = get_license_text(config, license_id, text, self.creation_time)
                    files.append(File(Path('LICENSES', f'{license_id}.txt'), license_text))

        if config['args']['cli']:
            from hatch.template import files_feature_cli

            files.extend(find_template_files(files_feature_cli))

        if self.plugin_config['tests']:
            from hatch.template import files_feature_tests

            files.extend(find_template_files(files_feature_tests))

        if self.plugin_config['ci']:
            from hatch.template import files_feature_ci

            files.extend(find_template_files(files_feature_ci))

        return files

    def finalize_files(self, config, files):
        if config['licenses']['headers'] and config['license_data']:
            for template_file in files:
                if template_file.path.name.endswith('.py'):
                    template_file.contents = config['license_header'] + template_file.contents

        if self.plugin_config['src-layout']:
            for template_file in files:
                if template_file.path.parts[0] == config['package_name']:
                    template_file.path = Path('src', template_file.path)


def get_license_text(config, license_id, license_text, creation_time):
    if license_id == 'MIT':
        license_text = license_text.replace('<year>', f'{creation_time.year}-present', 1)
        license_text = license_text.replace('<copyright holders>', f'{config["name"]} <{config["email"]}>', 1)
    elif license_id == 'BSD-3-Clause':
        license_text = license_text.replace('<year>', f'{creation_time.year}-present', 1)
        license_text = license_text.replace('<owner>', f'{config["name"]} <{config["email"]}>', 1)

    return f'{license_text.rstrip()}\n'



from __future__ import annotations

import os
import secrets
import stat
import sys
from errno import EACCES, EISDIR
from pathlib import Path


def raise_on_not_writable_file(filename: str) -> None:
    """
    Raise an exception if attempting to open the file for writing would fail.

    Separates files that can never be written from files that are writable but currently locked.

    :param filename: file to check

    :raises OSError: as if the file was opened for writing.

    """
    try:
        # lstat, not stat: it settles exists-and-writable in one syscall, and a hostile symlink planted at the lock
        # path would otherwise make this inspect the link target, letting an attacker turn a contended acquire into a
        # misleading PermissionError / IsADirectoryError and probe that target's attributes. The real open passes
        # O_NOFOLLOW and refuses the symlink anyway.
        file_stat = os.lstat(filename)
    except OSError:
        return  # does not exist, or an error the caller cannot act on

    # No mtime guard: the old `if st_mtime != 0` skip existed for NFS/Linux quirks where os.lstat could return an
    # all-zero struct, which it no longer does. Skipping on mtime 0 let a read-only file or a directory at the lock
    # path pass as missing, so acquire() blocked forever on an open that cannot succeed.
    if not (file_stat.st_mode & stat.S_IWUSR):
        raise PermissionError(EACCES, "Permission denied", filename)

    if stat.S_ISDIR(file_stat.st_mode):
        if sys.platform == "win32":  # pragma: win32 cover
            raise PermissionError(EACCES, "Permission denied", filename)
        raise IsADirectoryError(EISDIR, "Is a directory", filename)  # pragma: win32 no cover


def ensure_directory_exists(filename: Path | str) -> None:
    """
    Ensure the directory containing the file exists (create it if necessary).

    :param filename: file.

    """
    Path(filename).parent.mkdir(parents=True, exist_ok=True)


def break_lock_file(lock_file: str, mtime_before: float, ino_before: int) -> None:
    """
    Atomically break a stale lock file that was judged stale at modification time *mtime_before*.

    The file is renamed to a process-private name before being unlinked, so two processes breaking the same lock
    cannot delete each other's work (only one rename of a given inode succeeds; the loser gets ``OSError``). After the
    rename the file is re-checked: a newer modification time, or a different inode than *ino_before*, means a peer
    recreated the lock between the stale decision and the rename, so we grabbed a live file and must abort, leaving the
    renamed file in place rather than rolling back (a rollback rename is itself racy — same trade-off as the soft
    read/write marker break). The inode check matters because filesystems with coarse modification-time granularity
    (NFS, FAT) can give a same-second recreation the old mtime, so mtime alone would not catch it and a live lock would
    be unlinked; the inode is the reliable identity, mirroring the token re-check in the soft read/write marker break.
    ``lstat`` is used so a hostile symlink swapped in after the decision is not followed.

    The break name carries a random token so it is unguessable and unique per attempt. Without it two breakers in the
    same process share ``<lock>.break.<pid>``, and a second break can rename a freshly recreated live lock onto that
    path in the window between the re-verify ``lstat`` above and the ``unlink`` below, so we would delete a live lock
    the inode check just approved. A private name means nobody else can target our break path, matching the soft
    read/write marker break.

    :param lock_file: path to the lock file to break.
    :param mtime_before: modification time observed when the lock was judged stale.
    :param ino_before: inode number observed when the lock was judged stale.

    :raises OSError: if the rename fails (e.g. the file vanished or is not owned in a sticky directory).

    """
    break_path = f"{lock_file}.break.{os.getpid()}.{secrets.token_hex(16)}"
    Path(lock_file).rename(break_path)
    try:
        st_after = os.lstat(break_path)
    except OSError:
        return
    if st_after.st_mtime > mtime_before or st_after.st_ino != ino_before:
        return
    Path(break_path).unlink()


__all__ = [
    "break_lock_file",
    "ensure_directory_exists",
    "raise_on_not_writable_file",
]



from __future__ import annotations

from typing import Any


def get_default_config() -> dict[str, Any]:
    return {
        'installer': 'uv',
        'dependencies': [
            'coverage-enable-subprocess==1.0',
            'coverage[toml]~=7.4',
            'pytest~=8.1',
            'pytest-mock~=3.12',
            'pytest-randomly~=3.15',
            'pytest-rerunfailures~=14.0',
            'pytest-xdist[psutil]~=3.5',
        ],
        'scripts': {
            'run': 'pytest{env:HATCH_TEST_ARGS:} {args}',
            'run-cov': 'coverage run -m pytest{env:HATCH_TEST_ARGS:} {args}',
            'cov-combine': 'coverage combine',
            'cov-report': 'coverage report',
        },
        'matrix': [{'python': ['3.13', '3.12', '3.11', '3.10', '3.9', '3.8']}],
    }
